from re import I
from . import parsimony_pb2
import treeswift
from alive_progress import alive_it, alive_bar
from Bio import SeqIO

from dataclasses import dataclass
from collections import defaultdict


@dataclass
class AnnotatedMutation:
    genome_position: int  #0-based
    genome_residue: str
    cds: lambda: "CDS"
    codon_number: int  #0-based
    codon_start: int  #0-based
    codon_end: int  #0-based
    gene: str


def get_codon_table():
    bases = "TCAG"
    codons = [a + b + c for a in bases for b in bases for c in bases]
    amino_acids = 'FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'
    codon_table = dict(zip(codons, amino_acids))
    return codon_table


codon_table = get_codon_table()


def get_mutations(past_nuc_muts_dict, new_nuc_mutations_here, seq, cdses):

    annotated_mutations = []

    for mutation in new_nuc_mutations_here:
        cds = find_cds(mutation.position - 1, cdses)
        if cds:
            codon_number, codon_start, codon_end = find_codon(
                mutation.position - 1, cds)
            annotated_mutations.append(
                AnnotatedMutation(genome_position=mutation.position - 1,
                                  genome_residue=NUC_ENUM[mutation.mut_nuc[0]],
                                  gene=cds.qualifiers["gene"][0],
                                  codon_number=codon_number,
                                  codon_start=codon_start,
                                  codon_end=codon_end,
                                  cds=cds))

    by_gene_codon = defaultdict(list)

    for mutation in annotated_mutations:
        by_gene_codon[(mutation.gene, mutation.codon_number,
                       mutation.codon_start,
                       mutation.codon_end)].append(mutation)

    mutations_here = []
    for gene_codon, mutations in by_gene_codon.items():
        gene, codon_number, codon_start, codon_end = gene_codon
        very_initial_codon = seq[codon_start:codon_end]
        initial_codon = list(very_initial_codon)

        relevant_past_muts = [(x, past_nuc_muts_dict[x])
                              for x in range(codon_start, codon_end)
                              if x in past_nuc_muts_dict]
        for position, value in relevant_past_muts:
            initial_codon[position - codon_start] = value

        final_codon = initial_codon.copy()

        for mutation in mutations:
            pos_in_codon = mutation.genome_position - codon_start
            final_codon[pos_in_codon] = mutation.genome_residue

        initial_codon = "".join(initial_codon)
        final_codon = "".join(final_codon)
        initial_codon_trans = codon_table[initial_codon]
        final_codon_trans = codon_table[final_codon]
        if initial_codon_trans != final_codon_trans:
            mutations_here.append((gene, codon_number + 1, initial_codon_trans,
                                   final_codon_trans))

    # update past_nuc_muts_dict
    for mutation in annotated_mutations:
        past_nuc_muts_dict[mutation.genome_position] = mutation.genome_residue

    return mutations_here


def recursive_mutation_analysis(node, past_nuc_muts_dict, seq, cdses, pbar):
    pbar()
    new_nuc_mutations_here = node.nuc_mutations
    new_past_nuc_muts_dict = past_nuc_muts_dict.copy()
    node.aa_muts = get_mutations(new_past_nuc_muts_dict,
                                 new_nuc_mutations_here, seq, cdses)
    for child in node.children:
        recursive_mutation_analysis(child, new_past_nuc_muts_dict, seq, cdses,
                                    pbar)


NUC_ENUM = "ACGT"


def preorder_traversal(node):
    yield node
    for clade in node.children:
        yield from preorder_traversal(clade)


def preorder_traversal_iter(node):
    return iter(preorder_traversal(node))


def find_cds(position, cdses):
    for cds in cdses:
        if cds.location.start <= position <= cds.location.end:
            return cds
    return None


def find_codon(position, cds):
    # Get the codon number within the CDS
    codon_number = (position - cds.location.start) // 3
    codon_start = cds.location.start + codon_number * 3
    codon_end = codon_start + 3
    return codon_number, codon_start, codon_end


class UsherMutationAnnotatedTree:
    def __init__(self, tree_file, genbank_file=None):
        self.data = parsimony_pb2.data()
        self.data.ParseFromString(tree_file.read())
        self.condensed_nodes_dict = self.get_condensed_nodes_dict(
            self.data.condensed_nodes)
        print("Loading tree, this may take a while...")
        self.tree = treeswift.read_tree(self.data.newick, schema="newick")
        self.data.newick = ''

        self.annotate_mutations()

        self.expand_condensed_nodes()
        self.set_branch_lengths()
        if genbank_file:
            self.load_genbank_file(genbank_file)
            self.perform_aa_analysis()

    def perform_aa_analysis(self):
        seq = str(self.genbank.seq)
        with alive_bar(self.tree.num_nodes(),
                       title="Annotating amino acids") as pbar:
            recursive_mutation_analysis(self.tree.root, {}, seq, self.cdses,
                                        pbar)

    def load_genbank_file(self, genbank_file):
        self.genbank = SeqIO.read(genbank_file, "genbank")
        self.cdses = [x for x in self.genbank.features if x.type == "CDS"]
        # Assert that there are no compound locations and that all strands are positive,
        # and that all CDS features are a multiple of 3

        for cds in self.cdses:

            assert cds.location.strand == 1
            assert len(cds.location.parts) == 1
            assert len(cds.location.parts[0]) % 3 == 0

    def annotate_mutations(self):
        for i, node in alive_it(list(
                enumerate(preorder_traversal(self.tree.root))),
                                title="Annotating nuc muts"):
            node.nuc_mutations = self.data.node_mutations[i].mutation

    def set_branch_lengths(self):
        for i, node in alive_it(list(
                enumerate(preorder_traversal(self.tree.root))),
                                title="Setting branch length"):
            node.edge_length = len(node.nuc_mutations)

    def expand_condensed_nodes(self):
        for i, node in alive_it(list(enumerate(self.tree.traverse_leaves())),
                                title="Expanding condensed nodes"):

            if node.label and node.label in self.condensed_nodes_dict:

                for new_node_label in self.condensed_nodes_dict[node.label]:
                    new_node = treeswift.Node(label=new_node_label)
                    new_node.nuc_mutations = []
                    node.add_child(new_node)
                node.label = ""
            else:
                pass

    def get_condensed_nodes_dict(self, condensed_nodes_dict):
        output_dict = {}
        for condensed_node in alive_it(condensed_nodes_dict,
                                       title="Reading condensed nodes dict"):
            output_dict[
                condensed_node.node_name] = condensed_node.condensed_leaves
        return output_dict



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: parsimony.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='parsimony.proto',
  package='Parsimony',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0fparsimony.proto\x12\tParsimony\"^\n\x03mut\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0f\n\x07ref_nuc\x18\x02 \x01(\x05\x12\x0f\n\x07par_nuc\x18\x03 \x01(\x05\x12\x0f\n\x07mut_nuc\x18\x04 \x03(\x05\x12\x12\n\nchromosome\x18\x05 \x01(\t\"1\n\rmutation_list\x12 \n\x08mutation\x18\x01 \x03(\x0b\x32\x0e.Parsimony.mut\"=\n\x0e\x63ondensed_node\x12\x11\n\tnode_name\x18\x01 \x01(\t\x12\x18\n\x10\x63ondensed_leaves\x18\x02 \x03(\t\"*\n\rnode_metadata\x12\x19\n\x11\x63lade_annotations\x18\x01 \x03(\t\"\xa8\x01\n\x04\x64\x61ta\x12\x0e\n\x06newick\x18\x01 \x01(\t\x12\x30\n\x0enode_mutations\x18\x02 \x03(\x0b\x32\x18.Parsimony.mutation_list\x12\x32\n\x0f\x63ondensed_nodes\x18\x03 \x03(\x0b\x32\x19.Parsimony.condensed_node\x12*\n\x08metadata\x18\x04 \x03(\x0b\x32\x18.Parsimony.node_metadatab\x06proto3')
)




_MUT = _descriptor.Descriptor(
  name='mut',
  full_name='Parsimony.mut',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='Parsimony.mut.position', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ref_nuc', full_name='Parsimony.mut.ref_nuc', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='par_nuc', full_name='Parsimony.mut.par_nuc', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mut_nuc', full_name='Parsimony.mut.mut_nuc', index=3,
      number=4, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chromosome', full_name='Parsimony.mut.chromosome', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=124,
)


_MUTATION_LIST = _descriptor.Descriptor(
  name='mutation_list',
  full_name='Parsimony.mutation_list',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mutation', full_name='Parsimony.mutation_list.mutation', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=126,
  serialized_end=175,
)


_CONDENSED_NODE = _descriptor.Descriptor(
  name='condensed_node',
  full_name='Parsimony.condensed_node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_name', full_name='Parsimony.condensed_node.node_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='condensed_leaves', full_name='Parsimony.condensed_node.condensed_leaves', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=177,
  serialized_end=238,
)


_NODE_METADATA = _descriptor.Descriptor(
  name='node_metadata',
  full_name='Parsimony.node_metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clade_annotations', full_name='Parsimony.node_metadata.clade_annotations', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=240,
  serialized_end=282,
)


_DATA = _descriptor.Descriptor(
  name='data',
  full_name='Parsimony.data',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='newick', full_name='Parsimony.data.newick', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node_mutations', full_name='Parsimony.data.node_mutations', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='condensed_nodes', full_name='Parsimony.data.condensed_nodes', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='Parsimony.data.metadata', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=285,
  serialized_end=453,
)

_MUTATION_LIST.fields_by_name['mutation'].message_type = _MUT
_DATA.fields_by_name['node_mutations'].message_type = _MUTATION_LIST
_DATA.fields_by_name['condensed_nodes'].message_type = _CONDENSED_NODE
_DATA.fields_by_name['metadata'].message_type = _NODE_METADATA
DESCRIPTOR.message_types_by_name['mut'] = _MUT
DESCRIPTOR.message_types_by_name['mutation_list'] = _MUTATION_LIST
DESCRIPTOR.message_types_by_name['condensed_node'] = _CONDENSED_NODE
DESCRIPTOR.message_types_by_name['node_metadata'] = _NODE_METADATA
DESCRIPTOR.message_types_by_name['data'] = _DATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

mut = _reflection.GeneratedProtocolMessageType('mut', (_message.Message,), dict(
  DESCRIPTOR = _MUT,
  __module__ = 'parsimony_pb2'
  # @@protoc_insertion_point(class_scope:Parsimony.mut)
  ))
_sym_db.RegisterMessage(mut)

mutation_list = _reflection.GeneratedProtocolMessageType('mutation_list', (_message.Message,), dict(
  DESCRIPTOR = _MUTATION_LIST,
  __module__ = 'parsimony_pb2'
  # @@protoc_insertion_point(class_scope:Parsimony.mutation_list)
  ))
_sym_db.RegisterMessage(mutation_list)

condensed_node = _reflection.GeneratedProtocolMessageType('condensed_node', (_message.Message,), dict(
  DESCRIPTOR = _CONDENSED_NODE,
  __module__ = 'parsimony_pb2'
  # @@protoc_insertion_point(class_scope:Parsimony.condensed_node)
  ))
_sym_db.RegisterMessage(condensed_node)

node_metadata = _reflection.GeneratedProtocolMessageType('node_metadata', (_message.Message,), dict(
  DESCRIPTOR = _NODE_METADATA,
  __module__ = 'parsimony_pb2'
  # @@protoc_insertion_point(class_scope:Parsimony.node_metadata)
  ))
_sym_db.RegisterMessage(node_metadata)

data = _reflection.GeneratedProtocolMessageType('data', (_message.Message,), dict(
  DESCRIPTOR = _DATA,
  __module__ = 'parsimony_pb2'
  # @@protoc_insertion_point(class_scope:Parsimony.data)
  ))
_sym_db.RegisterMessage(data)


# @@protoc_insertion_point(module_scope)



#python usher_to_taxonium.py --input public-latest.all.masked.pb.gz --output ../taxonium_web_client/public/public2.jsonl.gz --metadata public-latest.metadata.tsv.gz --genbank hu1.gb --columns genbank_accession,country,date,pangolin_lineage

import json
import pandas as pd
from alive_progress import config_handler, alive_it, alive_bar
import treeswift
from . import ushertools
import argparse
import gzip

parser = argparse.ArgumentParser(
    description='Convert a Usher pb to Taxonium jsonl format')
parser.add_argument('--input',
                    type=str,
                    help='Input Usher pb file',
                    required=True)
parser.add_argument('--output',
                    type=str,
                    help='Output jsonl file',
                    required=True)
parser.add_argument('--metadata',
                    type=str,
                    help='Metadata file',
                    required=True)
parser.add_argument('--genbank', type=str, help='Genbank file', required=True)
parser.add_argument('--chronumental',
                    action='store_true',
                    help='If set, we will run chronumental')
parser.add_argument("--columns",
                    type=str,
                    help="Columns to include in the metadata")
parser.add_argument('--gzlevel', type=int, help='Gzip level', default=6)

args = parser.parse_args()

must_have_cols = ['strain']
cols_of_interest = set(args.columns.split(",")) if args.columns else set()
cols_of_interest.update(must_have_cols)
cols_of_interest = list(cols_of_interest)

#only load these cols:
print("Loading metadata file..")
metadata = pd.read_csv(args.metadata, sep="\t", usecols=cols_of_interest)
metadata.set_index("strain", inplace=True)
# convert metadata to dict of rows

metadata_dict = metadata.to_dict("index")
metadata_cols = metadata.columns
del metadata
print("Metadata loaded")

#config_handler.set_global(force_tty=True) # Uncommenting this will force progress bars

if "gz" in args.input:
    f = gzip.open(args.input, 'rb')
else:
    f = open(args.input, 'rb')

mat = ushertools.UsherMutationAnnotatedTree(f, args.genbank)

if args.chronumental:
    mat.tree.write_tree_newick("/tmp/distance_tree.nwk")

    print("Launching chronumental")
    import os

    os.system(
        "chronumental --tree /tmp/distance_tree.nwk --dates ./tfci.meta.tsv.gz --steps 140 --tree_out /tmp/timetree.nwk --dates_out ./date_comparison.tsv.gz"
    )

    # %%

    print("Reading time tree")
    time_tree = treeswift.read_tree("/tmp/timetree.nwk", schema="newick")
    time_tree_iter = ushertools.preorder_traversal(time_tree.root)
    for i, node in alive_it(enumerate(
            ushertools.preorder_traversal(mat.tree.root)),
                            title="Adding time tree"):
        time_tree_node = next(time_tree_iter)
        if args.chronumental:
            node.time_length = time_tree_node.edge_length
    del time_tree
    del time_tree_iter


def set_x_coords(root):
    """ Set x coordinates for the tree"""
    root.x_dist = 0
    root.x_time = 0
    for node in alive_it(root.traverse_preorder(),
                         title="Setting x coordinates"):
        if node.parent:
            node.x_dist = node.parent.x_dist + node.edge_length
            if args.chronumental:
                node.x_time = node.parent.x_time + node.time_length


def set_terminal_y_coords(root):
    for i, node in alive_it(enumerate(root.traverse_leaves()),
                            title="Setting terminal y coordinates"):
        node.y = i
        node.y = i


def set_internal_y_coords(root):
    # Each node should be halfway between the min and max y of its children
    for node in alive_it(root.traverse_postorder(leaves=False, internal=True),
                         title="Setting internal y coordinates"):

        child_ys = [child.y for child in node.children]
        node.y = (min(child_ys) + max(child_ys)) / 2


def get_all_aa_muts(root):
    all_aa_muts = set()
    for node in alive_it(list(root.traverse_preorder()),
                         title="Collecting all mutations"):
        if node.aa_muts:
            all_aa_muts.update(node.aa_muts)
    return list(all_aa_muts)


def make_aa_object(i, aa_tuple):
    # Tuple format is gene, position, prev, next
    gene, pos, prev, next = aa_tuple
    return {
        "gene": gene,
        "previous_residue": prev,
        "residue_pos": pos,
        "new_residue": next,
        "mutation_id": i,
    }


def get_node_object(node, node_to_index, metadata, aa_mut_tuple_to_index,
                    columns):

    object = {}
    object["name"] = node.label if node.label else ""
    # round to 5 dp
    object["x_dist"] = round(node.x_dist, 5)
    if args.chronumental:
        object["x_time"] = round(node.x_time, 5)
    object["y"] = node.y
    object['mutations'] = [
        aa_mut_tuple_to_index[aa_tuple] for aa_tuple in node.aa_muts
    ]
    # check if label is in metadata's index
    try:
        my_dict = metadata[node.label]
        for key in my_dict:
            value = my_dict[key]
            #if value is pd.NaN then set to empty string
            if pd.isna(value):
                value = ""
            object["meta_" + key] = value
    except KeyError:
        for key in columns:
            object["meta_" + key] = ""

    object['parent_id'] = node_to_index[
        node.parent] if node.parent else node_to_index[node]
    object['node_id'] = node_to_index[
        node]  # We don't strictly need this, but it doesn't add much to the space
    return object


print("Ladderizing tree..")
mat.tree.ladderize(ascending=False)
print("Ladderizing done")
set_x_coords(mat.tree.root)
set_terminal_y_coords(mat.tree.root)
set_internal_y_coords(mat.tree.root)

with alive_bar(title="Sorting on y") as bar:

    def return_y(node):
        bar()
        return node.y

    nodes_sorted_by_y = sorted(mat.tree.root.traverse_preorder(),
                               key=lambda x: return_y(x))

all_aa_muts_tuples = get_all_aa_muts(mat.tree.root)
all_aa_muts_objects = [
    make_aa_object(i, aa_tuple)
    for i, aa_tuple in alive_it(enumerate(all_aa_muts_tuples),
                                title="Making AA mut objects")
]
aa_mut_tuple_to_index = {
    aa_tuple: i
    for i, aa_tuple in enumerate(all_aa_muts_tuples)
}
first_json = {
    "aa_mutations": all_aa_muts_objects,
    "total_nodes": len(nodes_sorted_by_y),
}
node_to_index = {node: i for i, node in enumerate(nodes_sorted_by_y)}

if "gz" in args.output:
    output_file = gzip.open(args.output, 'wt', compresslevel=args.gzlevel)
else:
    output_file = open(args.output, 'wt')
output_file.write(json.dumps(first_json, separators=(',', ':')) + "\n")
for node in alive_it(nodes_sorted_by_y,
                     title="Converting each node, and writing out in JSON"):
    node_object = get_node_object(node, node_to_index, metadata_dict,
                                  aa_mut_tuple_to_index, metadata_cols)
    output_file.write(json.dumps(node_object) + "\n")

print("Done")

def main():
    # Ugly hack, fix it
    pass
